package org.psesquared.server.authentication.api.controller;

import jakarta.servlet.http.HttpServletResponse;
import java.util.List;
import lombok.NonNull;
import lombok.RequiredArgsConstructor;
import org.psesquared.server.authentication.api.service.AuthenticationService;
import org.psesquared.server.config.EmailConfigProperties;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.web.bind.annotation.DeleteMapping;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.PutMapping;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestParam;
import org.springframework.web.bind.annotation.RestController;

/**
 * This is a controller class for the Authentication API that handles the
 * requests from the client concerning login/logout and user account management.
 */
@RequestMapping("/api/2")
@RestController
@RequiredArgsConstructor
public class AuthenticationController {

  /**
   * The name of the HTTP location header.
   */
  private static final String LOCATION_HEADER = "Location";

  /**
   * The service class that this controller calls to further process requests.
   */
  private final AuthenticationService authenticationService;

  /**
   * The properties class that is used to return some externally stored URLs.
   */
  private final EmailConfigProperties emailConfigProperties;

  /**
   * The API-endpoint for registering a new
   * {@link org.psesquared.server.model.User} with a username, email address and
   * password. In order for the account to be used, the registration process
   * must be concluded with the verification of the email address. For this an
   * email with a link for verification is sent to {@code userInfo.email()}.
   *
   * @param userInfo The request-wrapper containing username, email and
   *                 password.
   * @return {@link HttpStatus#OK} on success, <br>
   *        {@link HttpStatus#BAD_REQUEST} for invalid user information
   * @see AuthenticationService#registerUser(UserInfoRequest)
   */
  @PostMapping("/auth/register.json")
  public ResponseEntity<String> registerUser(
      @RequestBody final UserInfoRequest userInfo) {
    return new ResponseEntity<>(authenticationService.registerUser(userInfo));
  }

  /**
   * The API-endpoint for verifying a newly created
   * {@link org.psesquared.server.model.User}. This method is invoked via the
   * link in the verification email that is sent in
   * {@link #registerUser(UserInfoRequest)}.
   * On success, it transfers the user to the dashboard and on failure it sets
   * the following status codes:
   * {@link HttpStatus#OK} on success, <br>
   * {@link HttpStatus#BAD_REQUEST} user exists and is already verified,
   * <br>
   * {@link HttpStatus#UNAUTHORIZED} invalid token, <br>
   * {@link HttpStatus#NOT_FOUND} user not found
   *
   * @param username The username of the user that needs to be verified
   * @param token    The JWT that indicates the authority of the request
   * @param response The {@link HttpServletResponse} for setting up a
   *                 redirection to the frontend
   * @see AuthenticationService#verifyRegistration(String, String)
   */
  @GetMapping("/auth/{username}/verify.json")
  public void verifyRegistration(
      @PathVariable final String username,
      @RequestParam("token") final String token,
      @NonNull final HttpServletResponse response) {
    HttpStatus status
        = authenticationService.verifyRegistration(username, token);
    if (status.equals(HttpStatus.OK)) {
      response.setHeader(LOCATION_HEADER,
          emailConfigProperties.dashboardBaseUrl());
      response.setStatus(HttpStatus.FOUND.value());
    } else {
      response.setStatus(status.value());
    }
  }

  /**
   * The API-endpoint for setting a JWT access token with a lifespan of one hour
   * as the "sessionid" cookie for authorization with further requests. <br>
   * (This is a secured endpoint requiring authorization via HTTP basic or JWT.)
   *
   * @param username The username of the user who wants to log in
   * @param response The {@link HttpServletResponse} for setting the "sessionid"
   *                 cookie
   * @return {@link HttpStatus#OK} on success, <br>
   *        {@link HttpStatus#NOT_FOUND} user not found
   * @see AuthenticationService#login(String, HttpServletResponse)
   */
  @PostMapping("/auth/{username}/login.json")
  public ResponseEntity<String> login(
      @PathVariable final String username,
      @NonNull final HttpServletResponse response) {
    return new ResponseEntity<>(
        authenticationService.login(username, response));
  }

  /**
   * The API-endpoint for invalidating the "sessionid" cookie containing a JWT
   * access token. Following authorized requests require HTTP basic
   * authentication or a new login. <br>
   * (This is a secured endpoint requiring authorization via HTTP basic or JWT.)
   *
   * @param username The username of the user who wants to log out
   * @param response The {@link HttpServletResponse} for invalidating the
   *                 "sessionid" cookie
   * @return {@link HttpStatus#OK} on success, <br>
   *        {@link HttpStatus#NOT_FOUND} user not found
   * @see AuthenticationService#logout(String, HttpServletResponse)
   */
  @PostMapping("/auth/{username}/logout.json")
  public ResponseEntity<String> logout(
      @PathVariable final String username,
      @NonNull final HttpServletResponse response) {
    return new ResponseEntity<>(
        authenticationService.logout(username, response));
  }

  /**
   * The API-endpoint for sending an email to the given address
   * ({@link ForgotPasswordRequest#email()} with an url to reset the password of
   * the user with that email address.
   *
   * @param email The email address of the user who wants to reset their
   *              password
   * @return {@link HttpStatus#OK} on success, <br>
   *        {@link HttpStatus#NOT_FOUND} user not found
   * @see AuthenticationService#forgotPassword(String)
   */
  @PostMapping("/auth/{email}/forgot.json")
  public ResponseEntity<String> forgotPassword(
      @PathVariable final String email) {
    return new ResponseEntity<>(authenticationService.forgotPassword(email));
  }

  /**
   * The API-endpoint for resetting the password of a
   * {@link org.psesquared.server.model.User}. This method is invoked via the
   * link in the verification email that is sent in
   * {@link #forgotPassword(String)}.
   *
   * @param username    The username of the user who wants to reset their
   *                    password
   * @param token       The JWT that indicates the authority of the request
   * @param requestBody The request-wrapper containing the new password
   * @return {@link HttpStatus#OK} on success, <br>
   *        {@link HttpStatus#BAD_REQUEST} password doesn't meet requirements,
   *        <br>
   *        {@link HttpStatus#UNAUTHORIZED} invalid token, <br>
   *        {@link HttpStatus#NOT_FOUND} user not found
   * @see AuthenticationService#resetPassword(String, String, PasswordRequest)
   */
  @PutMapping("/auth/{username}/resetpassword.json")
  public ResponseEntity<String> resetPassword(
      @PathVariable final String username,
      @RequestParam("token") final String token,
      @RequestBody final PasswordRequest requestBody) {
    return new ResponseEntity<>(
        authenticationService.resetPassword(username, token, requestBody));
  }

  /**
   * The API-endpoint for changing the password of a
   * {@link org.psesquared.server.model.User}, who is logged-in in the
   * dashboard.
   * (This is a secured endpoint requiring authorization via HTTP basic or JWT.)
   *
   * @param username The username of the user who wants to change their
   *                 password
   * @param requestBody The request-wrapper containing old and new password
   * @return {@link HttpStatus#OK} on success, <br>
   *        {@link HttpStatus#BAD_REQUEST} old password is wrong, <br>
   *        {@link HttpStatus#NOT_FOUND} user not found
   * @see AuthenticationService#changePassword(String, ChangePasswordRequest)
   */
  @PutMapping("/auth/{username}/changepassword.json")
  public ResponseEntity<String> changePassword(
      @PathVariable final String username,
      @RequestBody final ChangePasswordRequest requestBody) {
    return new ResponseEntity<>(
        authenticationService.changePassword(username, requestBody));
  }

  /**
   * The API-endpoint for deleting a {@link org.psesquared.server.model.User}.
   * This action is performed by a logged-in user from the dashboard.
   * The user must enter their password ({@link PasswordRequest#password()})
   * and if correct, the user along with all associated data is deleted.
   * (This is a secured endpoint requiring authorization via HTTP basic or JWT.)
   *
   * @param username The username of the user who wants to delete their
   *                 account
   * @param requestBody The request-wrapper containing the user's password
   * @return {@link HttpStatus#OK} on success, <br>
   *        {@link HttpStatus#BAD_REQUEST} wrong password, <br>
   *        {@link HttpStatus#NOT_FOUND} user not found
   * @see AuthenticationService#deleteUser(String, PasswordRequest)
   */
  @DeleteMapping("/auth/{username}/delete.json")
  public ResponseEntity<String> deleteUser(
      @PathVariable final String username,
      @RequestBody final PasswordRequest requestBody) {
    return new ResponseEntity<>(
        authenticationService.deleteUser(username, requestBody));
  }

  /**
   * This API-endpoint exists for compatibility with podcatchers, especially
   * AntennaPod and Kasts, which initially call this endpoint instead of
   * {@link #login(String, HttpServletResponse)}.
   * Accordingly, a call to this endpoint is internally treated as a login.
   * In particular, devices remain unsupported.
   *
   * @param username The username of the user to be synchronized
   * @param response The {@link HttpServletResponse} for setting the "sessionid"
   *                 cookie
   * @return A dummy response with a single dummy device for the given user
   * @see AuthenticationService#login(String, HttpServletResponse)
   */
  @GetMapping("/devices/{username}.json")
  public ResponseEntity<List<DeviceWrapper>> getDeviceList(
      @PathVariable final String username,
      @NonNull final HttpServletResponse response) {
    DeviceWrapper dummyDevice = new DeviceWrapper();
    return new ResponseEntity<>(
        List.of(dummyDevice),
        authenticationService.login(username, response));
  }

}
