package org.psesquared.server.authentication.api.data.access;

import java.util.Optional;
import org.psesquared.server.model.User;
import org.springframework.data.jpa.repository.JpaRepository;
import org.springframework.stereotype.Repository;

/**
 * This JPA repository manages all database transactions by automatically
 * implementing the logic behind custom queries via method naming convention.
 */
@Repository
public interface AuthenticationDao extends JpaRepository<User, Long> {

  /**
   * Checks if a user exists via their username.
   *
   * @param username The username
   * @return {@code true} if the user with the given username exists, <br>
   *        {@code false} otherwise
   */
  boolean existsByUsername(String username);

  /**
   * Finds the {@link User} with the given username if present.
   *
   * @param username The username of the user that is being searched for
   * @return An {@link Optional} containing the user with the given
   *        username if present
   */
  Optional<User> findByUsername(String username);

  /**
   * Finds the {@link User} with the given email address if present.
   *
   * @param email The email address of the user that is being searched for
   * @return An {@link Optional} containing the user with the given email
   *        address if present
   */
  Optional<User> findByEmail(String email);

  /**
   * Finds a {@link User} with the given username if present or with the
   * given email address otherwise.
   *
   * @param username The username of the user that is being searched for
   * @param email The email address of the user that is being searched for
   * @return An {@link Optional} containing the user with the given username
   *        or email address if present
   */
  Optional<User> findByUsernameOrEmail(String username, String email);

  /**
   * Deletes all users that haven't been verified yet and have registered
   * before the time specified by the given timestamp.
   *
   * @param timestamp The timestamp representing the number of seconds from
   *                  the epoch of 1970-01-01T00:00:00Z.
   */
  void deleteAllByEnabledFalseAndCreatedAtLessThan(long timestamp);

}
