package org.psesquared.server.authentication.api.service;

import lombok.RequiredArgsConstructor;
import org.psesquared.server.config.EmailConfigProperties;
import org.psesquared.server.config.JwtService;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.mail.SimpleMailMessage;
import org.springframework.mail.javamail.JavaMailSender;
import org.springframework.security.core.userdetails.UserDetails;
import org.springframework.stereotype.Service;

/**
 * This service class is responsible for sending emails to
 * {@link org.springframework.security.core.userdetails.User}s.
 *
 * @see JavaMailSender
 */
@Service
@RequiredArgsConstructor
public class EmailServiceImpl {

  /**
   * The {@link JavaMailSender} used for the sending of emails.
   */
  private final JavaMailSender emailSender;

  /**
   * The properties class that is used to return some externally stored URLs.
   */
  private final EmailConfigProperties emailConfigProperties;

  /**
   * The service class for managing the JWTs that are sent via email.
   */
  private final JwtService jwtService;

  /**
   * The email address from which the emails are sent.
   */
  @Value("${spring.mail.username}")
  private String sender;

  /**
   * The subject of the email that is sent for account verification.
   */
  private static final String VERIFICATION_MAIL_SUBJECT
      = "Bestätige deine E-Mail-Adresse für unseren"
      + " Podcast-Synchronisations-Server | Validate your Mail";

  /**
   * The subject of the email that is sent for resetting the password of a user.
   */
  private static final String PASSWORD_RESET_MAIL_SUBJECT
      = "Setze dein Passwort für unseren Podcast-Synchronisation-Server"
      + " zurück! | Reset Password";

  /**
   * The placeholder for the username.
   */
  private static final String USERNAME_MAIL_PLACEHOLDER = "username";

  /**
   * The placeholder for the verification URL.
   */
  private static final String VERIFICATION_MAIL_PLACEHOLDER
      = "verificationURL";

  /**
   * The placeholder for the URL for resetting the password of a user.
   */
  private static final String PASSWORD_RESET_MAIL_PLACEHOLDER
      = "passwordResetURL";

  /**
   * The question mark symbol announcing a URL query parameter.
   */
  private static final String URL_QUERY_PARAM = "?";

  /**
   * The format of the username URL query parameter.
   */
  private static final String USERNAME_PARAM = "username=";

  /**
   * The separator for URL query parameters.
   */
  private static final String PARAM_SEPARATOR = "&";

  /**
   * The format of the token URL query parameter.
   */
  private static final String TOKEN_PARAM = "token=";

  /**
   * The contents of the verification URL with placeholders read from an
   * external file.
   */
  @Value("#{T(org.psesquared.server.authentication.api.service"
      + ".ResourceReader).readFileToString('VerificationMail.txt')}")
  private String verificationMailText;

  /**
   * The contents of the URL for resetting the password of a user with
   * placeholders read from an external file.
   */
  @Value("#{T(org.psesquared.server.authentication.api.service"
      + ".ResourceReader).readFileToString('PasswordResetMail.txt')}")
  private String passwordResetMailText;

  /**
   * Sends a generic email to a user enabling him/her to perform a certain
   * action when clicking on the contained url.
   * This method uses a template which lies at resources and contains a
   * "verificationURL"-placeholder, which is replaced by the url.
   *
   * @param to Recipients email address
   * @param mailSubject Subject of the email
   * @param body Body of the email
   */
  private void sendMail(final String to,
                        final String mailSubject,
                        final String body) {
    // send simple mail message with credential from application.properties
    SimpleMailMessage message = new SimpleMailMessage();
    message.setFrom(sender);
    message.setTo(to);
    message.setSubject(mailSubject);
    message.setText(body);
    emailSender.send(message);
  }

  /**
   * Substitutes username and URL placeholders in email template.
   *
   * @param template The email template with placeholders
   * @param user The name of the user
   * @param url The URL with the JWT for request authentication
   * @return The email text with the actual username and URL
   */
  private String substitutePlaceholders(final String template,
                                        final UserDetails user,
                                        final String url) {
    return template
        .replace(USERNAME_MAIL_PLACEHOLDER, user.getUsername())
        .replace(VERIFICATION_MAIL_PLACEHOLDER, url)
        .replace(PASSWORD_RESET_MAIL_PLACEHOLDER, url);
  }

  /**
   * Generates the URL for verifying the account of a
   * {@link org.springframework.security.core.userdetails.User} containing
   * a JWT for authentication.
   *
   * @param userDetails The user details of the user who wants to verify their
   *                    account
   * @return The URL for verifying the user's account
   */
  private String generateVerificationUrlString(final UserDetails userDetails) {
    String token = jwtService.generateUrlTokenString(userDetails);
    String verificationUrl
        = String.format(emailConfigProperties.verificationUrl(),
                        userDetails.getUsername());

    return verificationUrl + URL_QUERY_PARAM + TOKEN_PARAM + token;
  }

  /**
   * Generates the URL for resetting the password of a
   * {@link org.springframework.security.core.userdetails.User} containing
   * a JWT for authentication.
   *
   * @param userDetails The user details of the user who wants to reset their
   *                    password
   * @return The URL for resetting the user's password
   */
  private String generatePasswordResetUrlString(final UserDetails userDetails) {
    final String token = jwtService.generateUrlTokenString(userDetails);
    return emailConfigProperties.dashboardBaseUrl()
            + emailConfigProperties.resetUrlPath()
            + URL_QUERY_PARAM
            + USERNAME_PARAM
            + userDetails.getUsername()
            + PARAM_SEPARATOR
            + TOKEN_PARAM
            + token;
  }

  /**
   * Sends a validation E-Mail to validate a user account by clicking on the
   * given URL.
   * It uses a template which lies at resources/ValidationMail.txt and contains
   * a "validationURL"-placeholder.
   *
   * @param to The email address of the user who wants to verify their account
   * @param userDetails The user details of that user
   */
  public void sendVerification(final String to, final UserDetails userDetails) {
    final String url = generateVerificationUrlString(userDetails);
    String mailText
        = substitutePlaceholders(verificationMailText, userDetails, url);

    sendMail(to, VERIFICATION_MAIL_SUBJECT, mailText);
  }

  /**
   * Sends a password-reset E-Mail to a user with a URL which lets the user
   * change his/her password.
   * It uses a template which lies at resources/PasswordResetMail.txt and
   * contains a "passwordResetURL"-placeholder.
   *
   * @param to The email address of the user who wants to reset their password
   * @param userDetails The user details of that user
   */
  public void sendPasswordReset(final String to,
                                final UserDetails userDetails) {
    final String url = generatePasswordResetUrlString(userDetails);
    String mailText
        = substitutePlaceholders(passwordResetMailText, userDetails, url);

    sendMail(to, PASSWORD_RESET_MAIL_SUBJECT, mailText);
  }
}
