package org.psesquared.server.authentication.api.service;

import jakarta.mail.internet.AddressException;
import jakarta.mail.internet.InternetAddress;
import java.util.regex.Pattern;
import lombok.RequiredArgsConstructor;
import org.springframework.stereotype.Service;

/**
 * The service class responsible for checking if user information (i.e.
 * username, email address and password) meets the specified requirements.
 */
@Service
@RequiredArgsConstructor
public class InputCheckService {

  /**
   * The strict boolean for
   * {@link InternetAddress#InternetAddress(String, boolean)}.
   */
  private static final boolean STRICT = true;

  /**
   * The return value for valid user information.
   */
  private static final boolean VALID = false;

  /**
   * The return value for invalid user information.
   */
  private static final boolean INVALID = true;

  /**
   * Asserts position at start of a line.
   */
  private static final String REGEX_START = "^";

  /**
   * Matches any word character (equivalent to [a-zA-Z0-9_]) character '-'
   * between 1 and 255 times.
   */
  private static final String USERNAME_REGEX_GROUP = "[\\w\\u002d]{1,255}";

  /**
   * Asserts that the password contains at least one digit.
   */
  private static final String PW_REGEX_GROUP1 = "(?=.*\\d)";

  /**
   * Asserts that the password contains at least one lower case character.
   */
  private static final String PW_REGEX_GROUP2 = "(?=.*[a-z])";

  /**
   * Asserts that the password contains at least one upper case character.
   */
  private static final String PW_REGEX_GROUP3 = "(?=.*[A-Z])";

  /**
   * Asserts that the password contains at least one special character from
   * the list [€°§´] or the Punct script extension.
   */
  private static final String PW_REGEX_GROUP4 = "(?=.*[\\p{Punct}€°§´])";

  /**
   * Asserts that the password contains only word characters
   * (equivalent to [a-zA-Z0-9_]) and the special characters specified in
   * {@link InputCheckService#PW_REGEX_GROUP4}.
   */
  private static final String PW_REGEX_GROUP5 = "[\\w\\p{Punct}€°§´]{8,255}";

  /**
   * Asserts position at the end of a line.
   */
  private static final String REGEX_END = "$";


  /**
   * The complete regex for a valid username consisting of the following regex
   * groups:
   * <br>
   * {@link #REGEX_START}, {@link #USERNAME_REGEX_GROUP}, {@link #REGEX_END}.
   */
  private static final String USERNAME_REGEX = REGEX_START
      + USERNAME_REGEX_GROUP
      + REGEX_END;

  /**
   * The complete regex for a valid password consisting of the following regex
   * groups:
   * <br>
   * {@link #REGEX_START}, {@link #PW_REGEX_GROUP1}, {@link #PW_REGEX_GROUP2},
   * {@link #PW_REGEX_GROUP3}, {@link #PW_REGEX_GROUP4},
   * {@link #PW_REGEX_GROUP5}, {@link #REGEX_END}.
   */
  private static final String PW_REGEX = REGEX_START
      + PW_REGEX_GROUP1
      + PW_REGEX_GROUP2
      + PW_REGEX_GROUP3
      + PW_REGEX_GROUP4
      + PW_REGEX_GROUP5
      + REGEX_END;

  /**
   * Checks if the given {@code username} meets the following requirements:
   * <br>
   * - contains only word characters (equivalent to [a-zA-Z0-9_])
   * and the character '-'.
   * <br>
   * - is between 1 and 255 characters long.
   *
   * @param username The username that needs to be validated
   * @return {@code false} if the username meets the requirements,
   *        <br>
   *        {@code true} otherwise
   */
  public boolean checkUsernameInvalid(final String username) {
    return !Pattern
        .compile(USERNAME_REGEX)
        .matcher(username)
        .matches();
  }

  /**
   * Checks if the given email address conforms to the RFC822 standard using
   * {@link InternetAddress#validate()}.
   *
   * @param email The email address that needs to be validated
   * @return {@code false} if the username meets the requirements,
   *        <br>
   *        {@code true} otherwise
   */
  public boolean checkEmailInvalid(final String email) {
    try {
      InternetAddress internetAddress = new InternetAddress(email, STRICT);
      internetAddress.validate();
      return VALID;
    } catch (AddressException e) {
      return INVALID;
    }
  }

  /**
   * Checks if the given {@code password} meets the following requirements:
   * <br>
   * - contains at least one digit.
   * <br>
   * - contains at least one lower case character.
   * <br>
   * - contains at least one upper case character.
   * <br>
   * - contains at least one special character from the list [€°§´] or the
   * Punct script extension.
   * <br>
   * - contains only word characters (equivalent to [a-zA-Z0-9_]) and special
   * characters specified above.
   *
   * @param password The username that needs to be validated
   * @return {@code false} if the username meets the requirements,
   *        <br>
   *        {@code true} otherwise
   */
  public boolean checkPasswordInvalid(final String password) {
    return !Pattern
        .compile(PW_REGEX)
        .matcher(password)
        .matches();
  }

}
