package org.psesquared.server.config;

import lombok.RequiredArgsConstructor;
import org.psesquared.server.authentication.api.data.access.AuthenticationDao;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.lang.NonNull;
import org.springframework.scheduling.annotation.EnableAsync;
import org.springframework.scheduling.annotation.EnableScheduling;
import org.springframework.security.authentication.AuthenticationManager;
import org.springframework.security.authentication.AuthenticationProvider;
import org.springframework.security.authentication.dao.DaoAuthenticationProvider;
import org.springframework.security.config.annotation.authentication.configuration.AuthenticationConfiguration;
import org.springframework.security.core.userdetails.UserDetailsService;
import org.springframework.security.core.userdetails.UsernameNotFoundException;
import org.springframework.security.crypto.bcrypt.BCryptPasswordEncoder;
import org.springframework.security.crypto.password.PasswordEncoder;
import org.springframework.transaction.annotation.EnableTransactionManagement;
import org.springframework.web.servlet.config.annotation.CorsRegistry;
import org.springframework.web.servlet.config.annotation.InterceptorRegistry;
import org.springframework.web.servlet.config.annotation.WebMvcConfigurer;

/**
 * The application configuration class declaring several beans.
 */
@Configuration
@EnableScheduling
@EnableTransactionManagement
@EnableAsync
@RequiredArgsConstructor
public class ApplicationConfig implements WebMvcConfigurer {

  /**
   * The message passed on to {@link UsernameNotFoundException}.
   */
  private static final String USERNAME_NOT_FOUND
      = "No user with the given username was found.";

  /**
   * The JPA repository that handles user related database requests.
   */
  private final AuthenticationDao authenticationDao;

  /**
   * Returns a {@link UserDetailsService} bean for retrieving users via username
   * from the database.
   *
   * @return {@link UserDetailsService}
   */
  @Bean
  public UserDetailsService userDetailsService() {
    return username -> authenticationDao.findByUsername(username)
        .orElseThrow(() -> new UsernameNotFoundException(USERNAME_NOT_FOUND));
  }

  /**
   * Returns an {@link AuthenticationProvider} bean for authenticating
   * {@link org.springframework.security.core.userdetails.User}s with username
   * and password using {@link #userDetailsService()} and
   * {@link #passwordEncoder()}.
   *
   * @return {@link AuthenticationProvider}
   */
  @Bean
  public AuthenticationProvider authenticationProvider() {
    DaoAuthenticationProvider authProvider = new DaoAuthenticationProvider();
    authProvider.setUserDetailsService(userDetailsService());
    authProvider.setPasswordEncoder(passwordEncoder());
    return authProvider;
  }

  /**
   * Returns a {@link BCryptPasswordEncoder} bean for password encryption.
   *
   * @return {@link PasswordEncoder}
   */
  @Bean
  public PasswordEncoder passwordEncoder() {
    return new BCryptPasswordEncoder();
  }

  /**
   * Returns an {@link AuthenticationManager} bean for processing authentication
   * requests from the given {@link AuthenticationConfiguration}.
   *
   * @param config The application's authentication configuration
   * @return {@link AuthenticationManager}
   * @throws Exception When the authentication manager couldn't be retrieved
   *        from the given configuration
   */
  @Bean
  public AuthenticationManager authenticationManager(
      final AuthenticationConfiguration config) throws Exception {
    return config.getAuthenticationManager();
  }

  /**
   * Returns a {@link WebMvcConfigurer} bean with CORS enabled globally.
   *
   * @return {@link WebMvcConfigurer}
   */
  @Bean
  public WebMvcConfigurer corsConfigurer() {
    return new WebMvcConfigurer() {
      @Override
      public void addCorsMappings(@NonNull final CorsRegistry registry) {
        registry
            .addMapping("/**")
            .allowedOrigins("*")
            .allowedMethods("*");
      }
    };
  }

  /**
   * Registers an {@link AuthenticationValidatorInterceptor}.
   *
   * @param registry The {@link InterceptorRegistry}
   */
  @Override
  public void addInterceptors(final InterceptorRegistry registry) {
    registry.addInterceptor(new AuthenticationValidatorInterceptor());
  }

}
