package org.psesquared.server.config;

import jakarta.servlet.http.HttpServletRequest;
import jakarta.servlet.http.HttpServletResponse;
import java.util.Map;
import org.springframework.lang.NonNull;
import org.springframework.security.authentication.AbstractAuthenticationToken;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.web.servlet.HandlerInterceptor;
import org.springframework.web.servlet.HandlerMapping;

/**
 * This interceptor class intercepts requests between the DispatcherServlet and
 * the Controller (i.e. when already mapped to Controller method).
 * It checks if the currently authenticated
 * {@link org.psesquared.server.model.User} is the same user for whom the
 * request is sent.
 */
public class AuthenticationValidatorInterceptor implements HandlerInterceptor {

  /**
   * The return value for aborting the execution of the Controller method.
   */
  private static final boolean ABORT_EXECUTION = false;

  /**
   * The return value for resuming the execution of the Controller method.
   */
  private static final boolean RESUME_EXECUTION = true;

  /**
   * The name of the username URL path variable.
   */
  private static final String PATH_VARIABLE_USERNAME = "username";

  /**
   * The default name associated with authentication.
   */
  private static final String USERNAME_NO_AUTH = "anonymousUser";

  /**
   * Checks if the currently authenticated
   * {@link org.psesquared.server.model.User} is the same user specified in the
   * URL path variable of the request.
   *
   * @param request The {@link HttpServletRequest}
   * @param response The {@link HttpServletResponse}
   * @param handler The chosen handler
   * @return {@code true} if the users match,
   *        <br>
   *        {@code false} otherwise
   */
  @Override
  public boolean preHandle(@NonNull final HttpServletRequest request,
                           @NonNull final HttpServletResponse response,
                           @NonNull final Object handler) {

    final String usernamePathVariable = extractUsernamePathVariable(request);
    final AbstractAuthenticationToken auth
        = (AbstractAuthenticationToken) SecurityContextHolder.getContext()
        .getAuthentication();
    final String usernameAuthenticated;

    if (usernamePathVariable == null || auth == null) {
      return RESUME_EXECUTION;
    }

    usernameAuthenticated = auth.getName();
    if (usernameAuthenticated == null
        || usernameAuthenticated.equals(USERNAME_NO_AUTH)
        || usernameAuthenticated.equals(usernamePathVariable)) {
      return RESUME_EXECUTION;
    }

    return ABORT_EXECUTION;
  }

  /**
   * Extracts the username path variable from the {@link HttpServletRequest}.
   *
   * @param request The {@link HttpServletRequest}
   * @return The value of the username path variable
   */
  private String extractUsernamePathVariable(final HttpServletRequest request) {
    // returns HttpServletRequest attribute that contains the URI templates map,
    // mapping variable names to values
    final Map<String, String> pathVariables = (Map<String, String>) request
        .getAttribute(HandlerMapping.URI_TEMPLATE_VARIABLES_ATTRIBUTE);
    // this attribute is of type Map<String, String> per definition,
    // so no type checks are needed
    return (pathVariables != null)
        ? pathVariables.get(PATH_VARIABLE_USERNAME) : null;
  }

}
