package org.psesquared.server.episode.actions.api.controller;

import java.util.List;
import lombok.RequiredArgsConstructor;
import org.psesquared.server.episode.actions.api.service.EpisodeActionService;
import org.psesquared.server.util.UpdateUrlsWrapper;
import org.springframework.http.ResponseEntity;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestParam;
import org.springframework.web.bind.annotation.RestController;

/**
 * This is a controller class for the Episode Action API that
 * handles the requests from the client concerning the synchronization
 * of episodes between clients.
 * In the end an appropriate response is sent back to the user.
 */
@RequestMapping("/api/2/episodes/{username}.json")
@RestController
@RequiredArgsConstructor
public class EpisodeActionController {

  /**
   * The service class that this controller calls to further process requests.
   */
  private final EpisodeActionService episodeActionService;

  /**
   * Takes a list of EpisodeActionPosts of a user and adds them to the database.
   *
   * @param username           The username of the user uploading the
   *                           EpisodeActions
   * @param episodeActionPosts The list of EpisodeActionPosts to be uploaded
   * @return The exit status of the function
   */
  @PostMapping
  public ResponseEntity<UpdateUrlsWrapper> addEpisodeActions(
      @PathVariable final String username,
      @RequestBody final List<EpisodeActionPost> episodeActionPosts) {
    episodeActionService.addEpisodeActions(username, episodeActionPosts);
    return ResponseEntity.ok(new UpdateUrlsWrapper());
  }

  /**
   * Returns a list of all EpisodeActions a user has uploaded so far in the form
   * of an EpisodeActionGetResponse.
   *
   * @param username The username of the user whose EpisodeActions are requested
   * @return The exit status with a response body containing all requested
   *        EpisodeActions
   */
  @GetMapping
  public ResponseEntity<EpisodeActionGetResponse> getEpisodeActions(
      @PathVariable final String username) {
    EpisodeActionGetResponse responseBody
        = new EpisodeActionGetResponse(episodeActionService
        .getEpisodeActions(username));
    return ResponseEntity.ok(responseBody);
  }

  /**
   * Returns a list of EpisodeActions of a user for a given podcast in the form
   * of an EpisodeActionGetResponse.
   *
   * @param username   The username of the user whose EpisodeActions are
   *                   requested
   * @param podcastUrl The RSS-Feed URL of the podcast in question
   * @return The exit status with a response body containing all requested
   *        EpisodeActions
   */
  @GetMapping(params = {"podcast"})
  public ResponseEntity<EpisodeActionGetResponse> getEpisodeActionsOfPodcast(
      @PathVariable final String username,
      @RequestParam("podcastUrl") final String podcastUrl) {
    EpisodeActionGetResponse responseBody
        = new EpisodeActionGetResponse(episodeActionService
        .getEpisodeActionsOfPodcast(username, podcastUrl));
    return ResponseEntity.ok(responseBody);
  }

  /**
   * Returns a list of EpisodeActions of a user since a given timestamp in the
   * form of an EpisodeActionGetResponse.
   *
   * @param username The username of the user whose EpisodeActions are requested
   * @param since    The timestamp signifying how old the EpisodeActions are
   *                 allowed to be
   * @return The exit status with a response body containing all requested
   *        EpisodeActions
   */
  @GetMapping(params = {"since"})
  public ResponseEntity<EpisodeActionGetResponse> getEpisodeActionsSince(
      @PathVariable final String username,
      @RequestParam("since") final long since) {
    EpisodeActionGetResponse responseBody
        = new EpisodeActionGetResponse(episodeActionService
        .getEpisodeActionsSince(username, since));
    return ResponseEntity.ok(responseBody);
  }

  /**
   * Returns a list of EpisodeActions of a user for a given podcast, since a
   * given time in the form of an EpisodeActionGetResponse.
   *
   * @param username   The username of the user whose EpisodeActions are
   *                   requested
   * @param podcastUrl The RSS-Feed URL of the podcast in question
   * @param since      The timestamp signifying how old the EpisodeActions are
   *                   allowed to be
   * @return The exit status with a response body containing all requested
   *        EpisodeActions
   */
  @GetMapping(params = {"podcast", "since"})
  public ResponseEntity<EpisodeActionGetResponse>
      getEpisodeActionsOfPodcastSince(
          @PathVariable final String username,
          @RequestParam("podcastUrl") final String podcastUrl,
          @RequestParam("since") final long since) {
    EpisodeActionGetResponse responseBody
        = new EpisodeActionGetResponse(episodeActionService
        .getEpisodeActionsOfPodcastSince(username, podcastUrl, since));
    return ResponseEntity.ok(responseBody);
  }

}
