package org.psesquared.server.episode.actions.api.data.access;

import java.time.LocalDateTime;
import java.util.List;
import java.util.Optional;
import org.psesquared.server.model.Action;
import org.psesquared.server.model.EpisodeAction;
import org.psesquared.server.model.User;
import org.springframework.data.jpa.repository.JpaRepository;
import org.springframework.stereotype.Repository;

/**
 * A DAO interface responsible for transactions involving EpisodeActions.
 */
@Repository
public interface EpisodeActionDao extends JpaRepository<EpisodeAction, Long> {

  /**
   * Find all EpisodeActions a user has uploaded.
   *
   * @param username The username of the user who uploaded the EpisodeActions
   * @return The list of EpisodeActions regarding the user
   */
  List<EpisodeAction> findByUserUsername(String username);

  /**
   * Find all EpisodeActions of a user that concern a certain podcast identified
   * by its RSS-Feed URL.
   *
   * @param username The username of the user who uploaded the EpisodeActions
   * @param url      The RSS-Feed URL of the podcast in question
   * @return The list of EpisodeActions regarding the user and the given podcast
   */
  List<EpisodeAction> findByUserUsernameAndEpisodeSubscriptionUrl(
      String username,
      String url);

  /**
   * Deletes all EpisodeActions of a podcast for a user.
   *
   * @param username The username of the user
   * @param url The podcast URL
   */
  void deleteByUserUsernameAndEpisodeSubscriptionUrl(
      String username,
      String url);

  /**
   * Checks if an EpisodeAction of the specified action type for a given user
   * and episode already exists.
   *
   * @param user The user
   * @param url The episode URL
   * @param action The type of action
   * @return {@code true} if such an EpisodeAction exists,
   *        <br>
   *        {@code false} otherwise
   */
  boolean existsByUserAndEpisodeUrlAndAction(User user,
                                             String url,
                                             Action action);

  /**
   * Finds the EpisodeAction of the specified action type and of an Episode
   * for a User.
   *
   * @param user The user
   * @param url The episode URL
   * @param action The type of action
   * @return An {@link Optional} containing the EpisodeAction if present
   */
  Optional<EpisodeAction> findByUserAndEpisodeUrlAndAction(User user,
                                                           String url,
                                                           Action action);

  /**
   * Find all EpisodeActions of a user since a given timestamp.
   *
   * @param username  The username of the user
   * @param timestamp The timestamp signifying how old an EpisodeAction is
   *                  allowed
   *                  to be
   * @return A list containing all EpisodeActions not older than the timestamp
   */
  List<EpisodeAction> findByUserUsernameAndTimestampGreaterThanEqual(
      String username,
      LocalDateTime timestamp);

  /**
   * Find all EpisodeActions of a user since a given timestamp of a given
   * podcast.
   *
   * @param username  The username of the user
   * @param timestamp The timestamp signifying how old an EpisodeAction is
   *                  allowed to be
   * @param url       The RSS-Feed URL of the podcast whose EpisodeActions are
   *                  requested
   * @return A list containing all EpisodeActions of the given podcast not older
   *        than the timestamp
   */
  List<EpisodeAction>
      findByUserUsernameAndTimestampGreaterThanEqualAndEpisodeSubscriptionUrl(
          String username,
          LocalDateTime timestamp,
          String url);

}
