package org.psesquared.server.episode.actions.api.service;

import java.time.LocalDateTime;
import java.time.ZoneOffset;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import lombok.RequiredArgsConstructor;
import org.psesquared.server.authentication.api.data.access.AuthenticationDao;
import org.psesquared.server.episode.actions.api.controller.EpisodeActionPost;
import org.psesquared.server.episode.actions.api.data.access.EpisodeActionDao;
import org.psesquared.server.episode.actions.api.data.access.EpisodeDao;
import org.psesquared.server.model.Action;
import org.psesquared.server.model.Episode;
import org.psesquared.server.model.EpisodeAction;
import org.psesquared.server.model.Subscription;
import org.psesquared.server.model.SubscriptionAction;
import org.psesquared.server.model.User;
import org.psesquared.server.subscriptions.api.data.access.SubscriptionActionDao;
import org.psesquared.server.subscriptions.api.data.access.SubscriptionDao;
import org.psesquared.server.util.RssParser;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

/**
 * This service class manages all business logic associated with the
 * episode action API.
 * <br>
 * It is called from the
 * {@link
 * org.psesquared.server.episode.actions.api.controller.EpisodeActionController}
 * and passes on requests concerning data access mainly to the
 * {@link EpisodeDao} and {@link EpisodeActionDao}.
 */
@Service
@Transactional
@RequiredArgsConstructor
public class EpisodeActionService {

  /**
   * The nano of second default value for
   * {@link LocalDateTime#ofEpochSecond(long, int, ZoneOffset)}.
   */
  private static final int NANO_OF_SECOND_DEFAULT = 0;

  /**
   * The JPA repository that handles all episode action related database
   * requests.
   */
  private final EpisodeActionDao episodeActionDao;

  /**
   * The JPA repository that handles all episode related database requests.
   */
  private final EpisodeDao episodeDao;

  /**
   * The JPA repository that handles all user related database requests.
   */
  private final AuthenticationDao authenticationDao;

  /**
   * The JPA repository that handles all subscription related database requests.
   */
  private final SubscriptionDao subscriptionDao;

  /**
   * The JPA repository that handles all subscription action related database
   * requests.
   */
  private final SubscriptionActionDao subscriptionActionDao;

  /**
   * The class for asynchronously fetching data from RSS feeds.
   */
  private final RssParser rssParser;

  /**
   * A map of subscription that need to be fetched with the {@link RssParser}.
   */
  private final Map<String, Subscription> subscriptionsToFetch
      = new HashMap<>();

  /**
   * Takes a list of EpisodeActionPosts, converts them to EpisodeActions and
   * saves them to the database.
   *
   * @param username           The username of the user who uploads these
   *                           EpisodeActions
   * @param episodeActionPosts List of EpisodeActionPosts that were sent via the
   *                           POST request
   */
  public void addEpisodeActions(
      final String username,
      final List<EpisodeActionPost> episodeActionPosts) {
    User user = authenticationDao.findByUsername(username).orElseThrow();
    List<EpisodeActionPost> filteredEpisodeActionPosts
        = new ArrayList<>(filterNewestAction(episodeActionPosts));
    List<EpisodeAction> episodeActions
        = episodeActionPostsToEpisodeActions(user, filteredEpisodeActionPosts);
    addEpisodeActionsToDatabase(user, episodeActions);
    validateDummyEpisodes();
  }

  private Collection<EpisodeActionPost> filterNewestAction(
      final List<EpisodeActionPost> episodeActionPosts) {
    Map<String, EpisodeActionPost> relevantEpisodeActionPosts = new HashMap<>();
    for (EpisodeActionPost episodeActionPost : episodeActionPosts) {
      if (episodeActionPost.getEpisodeAction().getAction() != Action.PLAY) {
        continue;
      }
      String url = episodeActionPost.getEpisodeUrl();
      if (relevantEpisodeActionPosts.containsKey(url)) {
        EpisodeActionPost currentEpisodeActionPost
            = relevantEpisodeActionPosts.get(url);
        if (episodeActionPost.getEpisodeAction().getTimestamp()
            .isAfter(
                currentEpisodeActionPost.getEpisodeAction().getTimestamp())) {
          relevantEpisodeActionPosts.put(url, episodeActionPost);
        }
      } else {
        relevantEpisodeActionPosts.put(url, episodeActionPost);
      }
    }
    return relevantEpisodeActionPosts.values();
  }

  private List<EpisodeAction> episodeActionPostsToEpisodeActions(
      final User user,
      final List<EpisodeActionPost> episodeActionPosts) {
    List<EpisodeAction> episodeActions = new ArrayList<>();
    for (EpisodeActionPost episodeActionPost : episodeActionPosts) {
      if (episodeActionPost.getEpisodeAction().getAction() == Action.PLAY) {
        episodeActions.add(episodeActionPostToEpisodeAction(
            user,
            episodeActionPost));
      }
    }
    return episodeActions;
  }

  private EpisodeAction episodeActionPostToEpisodeAction(
      final User user,
      final EpisodeActionPost episodeActionPost) {
    EpisodeAction episodeAction = episodeActionPost.getEpisodeAction();
    episodeAction.setUser(user);
    // If Subscription does not exist, create dummy Subscription
    Subscription subscription = null;
    if (!subscriptionDao.existsByUrl(episodeActionPost.getPodcastUrl())) {
      subscription = new Subscription();
      subscription.setTimestamp(
          LocalDateTime.now().toEpochSecond(ZoneOffset.UTC));
      subscription.setUrl(episodeActionPost.getPodcastUrl());
      subscription = subscriptionDao.save(subscription);
      // create Subscription Action
      SubscriptionAction subscriptionAction = SubscriptionAction.builder()
          .user(user)
          .added(true)
          .subscription(
              subscriptionDao.findByUrl(
                  episodeActionPost.getPodcastUrl()).orElseThrow())
          .timestamp(LocalDateTime.now().toEpochSecond(ZoneOffset.UTC))
          .build();
      subscriptionActionDao.save(subscriptionAction);
    } else {
      subscription = subscriptionDao
          .findByUrl(episodeActionPost.getPodcastUrl()).orElseThrow();
    }
    Episode episode = getEpisodeFromDatabase(episodeActionPost);
    episodeAction.setEpisode(episode);
    subscription.addEpisode(episode);
    return episodeAction;
  }

  private Episode getEpisodeFromDatabase(
      final EpisodeActionPost episodeActionPost) {
    Episode episode;
    String episodeUrl = episodeActionPost.getEpisodeUrl();
    String episodeGuid = episodeActionPost.getGuid();
    // If guid is passed and a matching episode exists get it
    if (episodeGuid != null && episodeDao.existsByGuid(episodeGuid)) {
      episode = episodeDao.findByGuid(episodeGuid).orElseThrow();
    } else if (episodeDao.existsByUrl(episodeUrl)) {
      // No episode with matching guid found -> search by url
      episode = episodeDao.findByUrl(episodeUrl).orElseThrow();
      // If guid was passed, pass it along to the database
      if (episodeGuid != null) {
        episode.setGuid(episodeGuid);
        episodeDao.save(episode);
      }
    } else {
      // Episode does not exist, so construct a new one
      episode = createEpisode(episodeActionPost);
    }
    return episode;
  }

  private Episode createEpisode(final EpisodeActionPost episodeActionPost) {
    Episode episode = Episode.builder()
        .title(episodeActionPost.getTitle())
        .url(episodeActionPost.getEpisodeUrl())
        .total(episodeActionPost.getTotal())
        .subscription(subscriptionDao
            .findByUrl(episodeActionPost.getPodcastUrl()).orElseThrow())
        .build();
    if (episodeActionPost.getGuid() != null) {
      episode.setGuid(episodeActionPost.getGuid());
    }
    episodeDao.save(episode);
    Subscription subscription = episode.getSubscription();
    subscriptionsToFetch.put(subscription.getUrl(), subscription);
    return episode;
  }

  private void addEpisodeActionsToDatabase(
      final User user,
      final List<EpisodeAction> episodeActions) {
    for (EpisodeAction episodeAction : episodeActions) {
      addEpisodeActionToDatabase(user, episodeAction);
    }
  }

  private void addEpisodeActionToDatabase(
      final User user,
      final EpisodeAction episodeAction) {
    if (episodeActionDao.existsByUserAndEpisodeUrlAndAction(
        user,
        episodeAction.getEpisode().getUrl(),
        episodeAction.getAction())) {
      addNewestEpisodeActionToDatabase(user, episodeAction);
    } else {
      episodeActionDao.save(episodeAction);
    }
  }

  private void addNewestEpisodeActionToDatabase(
      final User user,
      final EpisodeAction episodeAction) {
    EpisodeAction oldEpisodeAction
        = episodeActionDao.findByUserAndEpisodeUrlAndAction(
        user,
        episodeAction.getEpisode().getUrl(),
        episodeAction.getAction()).orElseThrow();
    if (episodeAction.getTimestamp().isAfter(oldEpisodeAction.getTimestamp())) {
      episodeActionDao.delete(oldEpisodeAction);
      episodeActionDao.save(episodeAction);
    }
  }

  private void validateDummyEpisodes() {
    Collection<Subscription> subscriptions = subscriptionsToFetch.values();
    for (Subscription subscription : subscriptions) {
      rssParser.validate(subscription);
    }
    subscriptionsToFetch.clear();
  }

  /**
   * Gets all EpisodeActions of a user and converts them to EpisodeActionPosts
   * before returning them.
   *
   * @param username The username of the user whose EpisodeActions are requested
   * @return A list containing the requested EpisodeActions as
   *        EpisodeActionPosts
   */
  public List<EpisodeActionPost> getEpisodeActions(final String username) {
    List<EpisodeAction> episodeActions
        = episodeActionDao.findByUserUsername(username);
    return episodeActionsToEpisodeActionPosts(episodeActions);
  }

  /**
   * Gets all EpisodeActions of a user that correspond to a given podcast.
   * Returns the EpisodeActions after converting them to EpisodeActionPosts.
   *
   * @param username   The username of the user whose EpisodeActions are
   *                   requested
   * @param podcastUrl The RSS-Feed URL of the podcast
   * @return A list containing the requested EpisodeActions as
   *        EpisodeActionPosts
   */
  public List<EpisodeActionPost> getEpisodeActionsOfPodcast(
      final String username,
      final String podcastUrl) {
    List<EpisodeAction> episodeActions
        = episodeActionDao.findByUserUsernameAndEpisodeSubscriptionUrl(
            username,
            podcastUrl);
    return episodeActionsToEpisodeActionPosts(episodeActions);
  }

  /**
   * Gets all EpisodeActions of a user since a given timestamp and converts them
   * to EpisodeActionPosts before returning them.
   *
   * @param username The username of the user whose EpisodeActions are requested
   * @param since    the timestamp signifying how old the EpisodeActions are
   *                 allowed to be
   * @return A list containing the requested EpisodeActions as
   *        EpisodeActionPosts
   */
  public List<EpisodeActionPost> getEpisodeActionsSince(
      final String username,
      final long since) {
    LocalDateTime sinceTimestamp
        = LocalDateTime.ofEpochSecond(
            since,
            NANO_OF_SECOND_DEFAULT,
            ZoneOffset.UTC);
    List<EpisodeAction> episodeActions
        = episodeActionDao.findByUserUsernameAndTimestampGreaterThanEqual(
            username,
            sinceTimestamp);
    return episodeActionsToEpisodeActionPosts(episodeActions);
  }

  /**
   * Gets all EpisodeActions of a user concerning a certain podcast since a
   * given timestamp and converts them to EpisodeActionPosts before returning
   * them.
   *
   * @param username   The username of the user whose EpisodeActions are
   *                   requested
   * @param podcastUrl The RSS-Feed URL of the podcast
   * @param since      The timestamp signifying how old the EpisodeActions are
   *                   allowed to be
   * @return A list containing the requested EpisodeActions as
   *        EpisodeActionPosts
   */
  public List<EpisodeActionPost> getEpisodeActionsOfPodcastSince(
      final String username,
      final String podcastUrl,
      final long since) {
    LocalDateTime sinceTimestamp
        = LocalDateTime.ofEpochSecond(
            since,
            NANO_OF_SECOND_DEFAULT,
            ZoneOffset.UTC);
    List<EpisodeAction> episodeActions = episodeActionDao
      .findByUserUsernameAndTimestampGreaterThanEqualAndEpisodeSubscriptionUrl(
            username,
            sinceTimestamp,
            podcastUrl);
    return episodeActionsToEpisodeActionPosts(episodeActions);
  }

  private List<EpisodeActionPost> episodeActionsToEpisodeActionPosts(
      final List<EpisodeAction> episodeActions) {
    List<EpisodeActionPost> episodeActionPosts = new ArrayList<>();

    for (EpisodeAction episodeAction : episodeActions) {
      episodeActionPosts.add(episodeAction.toEpisodeActionPost());
    }

    return episodeActionPosts;
  }

}
