package org.psesquared.server.subscriptions.api.data.access;

import java.util.List;
import java.util.Optional;
import org.psesquared.server.model.Subscription;
import org.psesquared.server.model.SubscriptionAction;
import org.psesquared.server.model.User;
import org.springframework.data.jpa.repository.JpaRepository;
import org.springframework.stereotype.Repository;

/**
 * A DAO interface responsible for transactions involving SubscriptionActions.
 */
@Repository
public interface SubscriptionActionDao
    extends JpaRepository<SubscriptionAction, Long> {

  /**
   * True, if the given user is already subscribed to the given Subscription.
   *
   * @param user         The user that could be subscribed
   * @param subscription The subscription the user could be subscribed to
   * @return A boolean value signifying whether the user is subscribed to the
   *        given subscription
   */
  boolean existsByUserAndSubscription(User user, Subscription subscription);

  /**
   * Find the SubscriptionAction signifying that the user is subscribed to the
   * given Subscription.
   *
   * @param user         The user who is subscribed to the subscription
   * @param subscription The subscription that the user is subscribed to
   * @return Contains the relevant SubscriptionAction. Could also be NULL if
   *        none was found.
   */
  Optional<SubscriptionAction> findByUserAndSubscription(
      User user,
      Subscription subscription);

  /**
   * Find the SubscriptionAction for a {@link User} with the given username
   * and for a {@link Subscription} with the given URL.
   *
   * @param username         The username of the user who is subscribed to
   *                         the subscription
   * @param subscriptionUrl The URL of the subscription that the user is
   *                        subscribed to
   * @return Contains the relevant SubscriptionAction. Could also be NULL if
   *        none was found.
   */
  Optional<SubscriptionAction> findByUserUsernameAndSubscriptionUrl(
      String username, String subscriptionUrl);

  /**
   * All SubscriptionActions of a given user that were applied since a given
   * timestamp are searched for and returned.
   *
   * @param username  The username of the user whose SubscriptionActions are
   *                  requested
   * @param timestamp The timestamp signifying how old the SubscriptionActions
   *                  are allowed to be
   * @return A list of SubscriptionActions that have since been applied
   */
  List<SubscriptionAction> findByUserUsernameAndTimestampGreaterThanEqual(
      String username,
      long timestamp);

  /**
   * Returns a List of all Subscriptions the user is subscribed to.
   *
   * @param username The username of the user whose subscriptions are requested
   * @return A list of subscriptions the user is subscribed to
   */
  List<SubscriptionAction> findByUserUsernameAndAddedTrue(String username);

  /**
   * Returns a List of RSS-Feed URLs of all podcasts the given user is
   * subscribed to since a given timestamp.
   *
   * @param username  The username of the user whose subscriptions are requested
   * @param timestamp The timestamp signifying the time since when the user must
   *                  have been subscribed
   * @return List of RSS-Feed URLs
   */
  List<SubscriptionAction>
      findByUserUsernameAndAddedTrueAndTimestampGreaterThanEqual(
          String username,
          long timestamp);

}
